#include "stdafx.h"
#include "D3D11Renderer.hpp"

namespace
{
#define HCHECK(value) HCheck(#value, __LINE__, value)

void HCheck(const char* what, int line, HRESULT hr)
{
    if (FAILED(hr)) {
        printf("%s failed with code %ld at line %d\n", what, hr, line);
        abort();
    }
}

#pragma region Math

struct Matrix4x4f {
    float m[16];
};

DirectX::XMMATRIX XM_CALLCONV poseToXmMatrix(const XrPosef& pose)
{
    return XMMatrixAffineTransformation(DirectX::g_XMOne, DirectX::g_XMZero, XMLoadFloat4(reinterpret_cast<const DirectX::XMFLOAT4*>(&pose.orientation)),
        XMLoadFloat3(reinterpret_cast<const DirectX::XMFLOAT3*>(&pose.position)));
}

DirectX::XMMATRIX XM_CALLCONV LoadXrMatrix(const Matrix4x4f& matrix)
{
    // XrMatrix4x4f has same memory layout as DirectX Math (Row-major,post-multiplied = column-major,pre-multiplied)
    return XMLoadFloat4x4(reinterpret_cast<const DirectX::XMFLOAT4X4*>(&matrix));
}

void createProjection(
    Matrix4x4f* result, float tanAngleLeft, float tanAngleRight, float tanAngleUp, float const tanAngleDown, const float nearZ, const float farZ)
{
    const float tanAngleWidth = tanAngleRight - tanAngleLeft;
    const float tanAngleHeight = tanAngleUp - tanAngleDown;

    const float offsetZ = 0;

    result->m[0] = 2 / tanAngleWidth;
    result->m[4] = 0;
    result->m[8] = (tanAngleRight + tanAngleLeft) / tanAngleWidth;
    result->m[12] = 0;

    result->m[1] = 0;
    result->m[5] = 2 / tanAngleHeight;
    result->m[9] = (tanAngleUp + tanAngleDown) / tanAngleHeight;
    result->m[13] = 0;

    result->m[2] = 0;
    result->m[6] = 0;
    result->m[10] = -(farZ + offsetZ) / (farZ - nearZ);
    result->m[14] = -(farZ * (nearZ + offsetZ)) / (farZ - nearZ);

    result->m[3] = 0;
    result->m[7] = 0;
    result->m[11] = -1;
    result->m[15] = 0;
}

void createProjectionFov(Matrix4x4f* result, XrFovf fov, float nearZ, float farZ)
{
    const float tanLeft = tanf(fov.angleLeft);
    const float tanRight = tanf(fov.angleRight);

    const float tanDown = tanf(fov.angleDown);
    const float tanUp = tanf(fov.angleUp);

    createProjection(result, tanLeft, tanRight, tanUp, tanDown, nearZ, farZ);
}
#pragma endregion

ComPtr<IDXGIAdapter> getAdapter(LUID luid)
{
    ComPtr<IDXGIFactory> factory = nullptr;

    const HRESULT hr = CreateDXGIFactory1(IID_PPV_ARGS(&factory));
    if (SUCCEEDED(hr)) {
        UINT i = 0;
        while (true) {
            ComPtr<IDXGIAdapter> adapter = nullptr;
            if (factory->EnumAdapters(i++, &adapter) == DXGI_ERROR_NOT_FOUND) break;
            DXGI_ADAPTER_DESC desc;
            if (SUCCEEDED(adapter->GetDesc(&desc)) && desc.AdapterLuid.HighPart == luid.HighPart && desc.AdapterLuid.LowPart == luid.LowPart) {
                return adapter;
            }
        }
    }

    return nullptr;
}
}  // namespace


D3D11Renderer::D3D11Renderer(XrInstance instance, XrSystemId systemId, bool mrEnabled)
    : IRenderer(instance, systemId)
{
    PFN_xrGetD3D11GraphicsRequirementsKHR pfnGetD3D11GraphicsRequirementsKHR = nullptr;
    XRCHECK(xrGetInstanceProcAddr(instance, "xrGetD3D11GraphicsRequirementsKHR", reinterpret_cast<PFN_xrVoidFunction*>(&pfnGetD3D11GraphicsRequirementsKHR)));

    XrGraphicsRequirementsD3D11KHR graphicsRequirements{XR_TYPE_GRAPHICS_REQUIREMENTS_D3D11_KHR};
    XRCHECK(pfnGetD3D11GraphicsRequirementsKHR(instance, systemId, &graphicsRequirements));
    const ComPtr<IDXGIAdapter> adapter = getAdapter(graphicsRequirements.adapterLuid);

    const UINT deviceFlags = D3D11_CREATE_DEVICE_DEBUG;
    const auto type = adapter == nullptr ? D3D_DRIVER_TYPE_HARDWARE : D3D_DRIVER_TYPE_UNKNOWN;
    HCHECK(D3D11CreateDevice(adapter.Get(), type, nullptr, deviceFlags, nullptr, 0, D3D11_SDK_VERSION, &m_device, nullptr, &m_dc));

    m_cubeRot = 0;
    initCube( 0.0f, 0.0f, -1.0f, 1.0f);
    initCube(-1.5f, 0.0f, -3.0f, mrEnabled ? 0.6f : 1.0f);
    initShaders();
    initConstants();

    m_graphicsBindings.device = GetD3DDevice();

#ifdef SHOW_WINDOW
    m_window = std::make_unique<Window>(GetD3DDevice(), 2048, 1152);
#endif
}

void D3D11Renderer::insertColorSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) { enumerateSwapchainImages(handle, m_colorSwapchainImagesMap); }

void D3D11Renderer::insertDepthSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) { enumerateSwapchainImages(handle, m_depthSwapchainImagesMap); }

void D3D11Renderer::enumerateSwapchainImages(XrSwapchain handle, std::unordered_map<XrSwapchain, std::vector<XrSwapchainImageD3D11KHR>>& swapchainImagesMap)
{
    uint32_t imageCount;
    XRCHECK(xrEnumerateSwapchainImages(handle, 0, &imageCount, nullptr));

    std::vector<XrSwapchainImageD3D11KHR> swapchainImages(imageCount);
    for (uint32_t i = 0; i < imageCount; i++) {
        swapchainImages[i] = XrSwapchainImageD3D11KHR{XR_TYPE_SWAPCHAIN_IMAGE_D3D11_KHR};
    }
    XRCHECK(xrEnumerateSwapchainImages(handle, imageCount, &imageCount, reinterpret_cast<XrSwapchainImageBaseHeader*>(swapchainImages.data())));

    D3D11_TEXTURE2D_DESC desc{};
    swapchainImages[0].texture->GetDesc(&desc);

    swapchainImagesMap.insert(std::make_pair(handle, swapchainImages));
}

void D3D11Renderer::update(float deltaTime) { m_cubeRot += 0.35f * deltaTime; }

void D3D11Renderer::render(XrSwapchain colorHandle, XrSwapchain depthHandle, uint32_t colorIndex, uint32_t depthIndex, const XrCompositionLayerProjectionView& view)
{
    ID3D11Texture2D* texture = m_colorSwapchainImagesMap[colorHandle][colorIndex].texture;
    const ComPtr<ID3D11RenderTargetView> renderTarget = getRenderTarget(texture);
    const ComPtr<ID3D11DepthStencilView> depthView = getDepthMap(colorHandle, depthHandle, colorIndex, depthIndex);

    clear(renderTarget.Get(), depthView.Get());

    ID3D11RenderTargetView* tgts[]{renderTarget.Get()};
    m_dc->OMSetRenderTargets(1, tgts, depthView.Get());
    m_dc->RSSetState(m_rasterizerState.Get());
    m_dc->OMSetDepthStencilState(m_depthStencilState.Get(), 0);

    {
        D3D11_VIEWPORT d3dviewport{};
        d3dviewport.TopLeftX = static_cast<float>(view.subImage.imageRect.offset.x);
        d3dviewport.TopLeftY = static_cast<float>(view.subImage.imageRect.offset.y);
        d3dviewport.Width = static_cast<float>(view.subImage.imageRect.extent.width);
        d3dviewport.Height = static_cast<float>(view.subImage.imageRect.extent.height);
        d3dviewport.MinDepth = 0;
        d3dviewport.MaxDepth = 1;
        m_dc->RSSetViewports(1, &d3dviewport);
    }

    for (const Cube& cube : m_cubes) {
        // Update subresource
        {
            const DirectX::XMMATRIX viewMatrix = XMMatrixTranspose(XMMatrixInverse(nullptr, poseToXmMatrix(view.pose)));
            XMStoreFloat4x4(&m_constants.m_data.view, viewMatrix);

            Matrix4x4f projectionMatrix{};
            createProjectionFov(&projectionMatrix, view.fov, 0.05f, 100.0f);

            XMStoreFloat4x4(&m_constants.m_data.projection, XMMatrixTranspose(LoadXrMatrix(projectionMatrix)));
            XMStoreFloat4x4(&m_constants.m_data.world,
                XMMatrixTranspose(DirectX::XMMatrixRotationY(m_cubeRot) * DirectX::XMMatrixTranslation(cube.posX, cube.posY, cube.posZ)));
            m_dc->UpdateSubresource(m_constants.m_constantBuffer.Get(), 0, nullptr, &m_constants.m_data, 0, 0);
        }

        {
            UINT stride = cube.vertexStride;
            UINT offset = 0;
            ID3D11Buffer* bufs[]{cube.vertexBuffer.Get()};
            m_dc->IASetVertexBuffers(0, sizeof(bufs) / sizeof(*(bufs)), bufs, &stride, &offset);
        }
        m_dc->IASetIndexBuffer(cube.indexBuffer.Get(), DXGI_FORMAT_R16_UINT, 0);
        m_dc->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

        m_dc->IASetInputLayout(m_shaders.inputLayout.Get());
        m_dc->VSSetShader(m_shaders.vertexShader.Get(), nullptr, 0);
        m_dc->VSSetConstantBuffers(0, 1, m_constants.m_constantBuffer.GetAddressOf());
        m_dc->PSSetShader(m_shaders.pixelShader.Get(), nullptr, 0);

        m_dc->DrawIndexed(cube.indexCount, 0, 0);
    }

    ID3D11RenderTargetView* nullTgts[] = {nullptr};
    m_dc->OMSetRenderTargets(1, nullTgts, nullptr);
}

void D3D11Renderer::copyForWindow(XrSwapchain handle, int index)
{
#ifdef SHOW_WINDOW
    auto backBuffer = m_window->backBufferRtv();
    ComPtr<ID3D11DeviceContext> ctx;
    GetD3DDevice()->GetImmediateContext(ctx.GetAddressOf());
    const float color[]{1, 0.0f, 0.0f, 0.0f};
    ctx->ClearRenderTargetView(backBuffer.Get(), color);
    ComPtr<ID3D11Texture2D> backTexture;
    backBuffer->GetResource(reinterpret_cast<ID3D11Resource**>(backTexture.GetAddressOf()));
    ctx->CopyResource(backTexture.Get(), m_colorSwapchainImagesMap[handle][index].texture);
#endif
}

bool D3D11Renderer::showWindow()
{
#ifdef SHOW_WINDOW
    return m_window->runEventLoop();
#endif
    return true;
}

void D3D11Renderer::initCube(float posX, float posY, float posZ, float alpha)
{
    m_cubes.emplace_back();
    Cube& cube = m_cubes.back();

    cube.posX = posX;
    cube.posY = posY;
    cube.posZ = posZ;

    float vertices[][7]{
        {-0.1f, -0.1f, -0.1f, 0, 0, 0, alpha},
        {-0.1f, -0.1f,  0.1f, 0, 0, 1, alpha},
        {-0.1f,  0.1f, -0.1f, 0, 1, 0, alpha},
        {-0.1f,  0.1f,  0.1f, 0, 1, 1, alpha},
        { 0.1f, -0.1f, -0.1f, 1, 0, 0, alpha},
        { 0.1f, -0.1f,  0.1f, 1, 0, 1, alpha},
        { 0.1f,  0.1f, -0.1f, 1, 1, 0, alpha},
        { 0.1f,  0.1f,  0.1f, 1, 1, 1, alpha},
    };
    createBuf(vertices, sizeof(vertices), D3D11_BIND_VERTEX_BUFFER, &cube.vertexBuffer);
    cube.vertexStride = 7 * sizeof(float);
    unsigned short indices[][3]{
        {0, 2, 1},
        {1, 2, 3},
        {4, 5, 6},
        {5, 7, 6},
        {0, 1, 5},
        {0, 5, 4},
        {2, 6, 7},
        {2, 7, 3},
        {0, 4, 6},
        {0, 6, 2},
        {1, 3, 7},
        {1, 7, 5},
    };
    cube.indexCount = sizeof indices / sizeof(unsigned short);
    createBuf(indices, sizeof(indices), D3D11_BIND_INDEX_BUFFER, &cube.indexBuffer);
}

void D3D11Renderer::initConstants()
{
    CD3D11_BUFFER_DESC cbDesc(sizeof m_constants.m_data, D3D11_BIND_CONSTANT_BUFFER);
    HCHECK(m_device->CreateBuffer(&cbDesc, nullptr, m_constants.m_constantBuffer.GetAddressOf()));
}

void D3D11Renderer::initShaders()
{
    ComPtr<ID3DBlob> compiledVertexShader = compileShader(
        "cbuffer ConstantBuffer : register(b0) {"
        "  matrix world;"
        "  matrix view;"
        "  matrix projection;"
        "};"
        "struct VsInput {"
        "  float3 pos : POSITION;"
        "  float4 color : COLOR;"
        "};"
        "struct VsOutput {"
        "  float4 position : SV_POSITION;"
        "  float4 color : COLOR;"
        "};"
        "VsOutput main(VsInput input) {"
        "  VsOutput output;"
        "  float4 pos = float4(input.pos, 1.0f);"
        "  pos = mul(pos, world);"
        "  pos = mul(pos, view);"
        "  pos = mul(pos, projection);"
        "  output.position = pos;"
        "  output.color = float4(input.color.rgb * input.color.a, input.color.a);"
        "  return output;"
        "}",
        "vs_4_0");

    ComPtr<ID3DBlob> compiledPixelShader = compileShader(
        "struct PsInput {"
        "  float4 position : SV_POSITION;"
        "  float4 color : COLOR;"
        "};"
        "float4 main(PsInput input) : SV_TARGET {"
        "  return input.color;"
        "}",
        "ps_4_0");

    HCHECK(m_device->CreateVertexShader(
        compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), nullptr, m_shaders.vertexShader.GetAddressOf()));
    HCHECK(m_device->CreatePixelShader(compiledPixelShader->GetBufferPointer(), compiledPixelShader->GetBufferSize(), nullptr, &m_shaders.pixelShader));

    D3D11_INPUT_ELEMENT_DESC iaDesc[]{
        {"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0},
        {"COLOR", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0},
    };

    HCHECK(m_device->CreateInputLayout(
        iaDesc, sizeof(iaDesc) / sizeof(*(iaDesc)), compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), &m_shaders.inputLayout));
}

void D3D11Renderer::createBuf(void* bufdata, int datasize, UINT bindFlags, ID3D11Buffer** buf) const
{
    CD3D11_BUFFER_DESC vDesc(datasize, bindFlags);
    D3D11_SUBRESOURCE_DATA subresourceData{};
    subresourceData.pSysMem = bufdata;
    subresourceData.SysMemPitch = 0;
    subresourceData.SysMemSlicePitch = 0;
    HCHECK(m_device->CreateBuffer(&vDesc, &subresourceData, buf));
}

ComPtr<ID3DBlob> D3D11Renderer::compileShader(const char* src, const char* target)
{
    ComPtr<ID3DBlob> compiledShader = nullptr;
    ComPtr<ID3DBlob> compilerMsgs = nullptr;
    const HRESULT hr =
        D3DCompile(src, strlen(src) + 1, "context_vs", nullptr, nullptr, "main", target, D3DCOMPILE_OPTIMIZATION_LEVEL3, 0, &compiledShader, &compilerMsgs);
    if (compilerMsgs != nullptr)
        printf("compiler errors: '%.*s'\n", static_cast<int>(compilerMsgs->GetBufferSize()), reinterpret_cast<char*>(compilerMsgs->GetBufferPointer()));
    HCHECK(hr);
    return compiledShader;
}

void D3D11Renderer::clear(ID3D11RenderTargetView* colorRTV, ID3D11DepthStencilView* depthView) const
{
    static const float color[]{0.0f, 0.0f, 0.0f, 0.0f};
    m_dc->ClearRenderTargetView(colorRTV, color);
    if (depthView) m_dc->ClearDepthStencilView(depthView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);
}

ComPtr<ID3D11DepthStencilView> D3D11Renderer::getDepthMap(XrSwapchain colorSwapchain, XrSwapchain depthSwapchain, uint32_t colorIndex, uint32_t depthIndex)
{
    const XrSwapchain swapchainKey = depthSwapchain == XR_NULL_HANDLE ? colorSwapchain : depthSwapchain;
    const uint32_t swapchainIndex = depthSwapchain == XR_NULL_HANDLE ? colorIndex : depthIndex;
    const std::pair<XrSwapchain, uint32_t> key = std::make_pair(swapchainKey, swapchainIndex);
    const auto iter = m_depthMap.find(key);
    if (iter != m_depthMap.end()) {
        return iter->second;
    }

    ComPtr<ID3D11Texture2D> depthTexture{};

    if (depthSwapchain != XR_NULL_HANDLE) {
        depthTexture = m_depthSwapchainImagesMap[depthSwapchain][depthIndex].texture;
    } else {
        D3D11_TEXTURE2D_DESC desc{};
        m_colorSwapchainImagesMap[colorSwapchain][colorIndex].texture->GetDesc(&desc);
        CD3D11_TEXTURE2D_DESC depthStencilDesc(
            DXGI_FORMAT_D24_UNORM_S8_UINT, static_cast<UINT>(desc.Width), static_cast<UINT>(desc.Height), 1, 1, D3D11_BIND_DEPTH_STENCIL);

        m_device->CreateTexture2D(&depthStencilDesc, nullptr, &depthTexture);
    }

    D3D11_TEXTURE2D_DESC desc{};
    depthTexture->GetDesc(&desc);
    CD3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc(D3D11_DSV_DIMENSION_TEXTURE2D, DXGI_FORMAT_D24_UNORM_S8_UINT);
    ComPtr<ID3D11DepthStencilView> depthStencilView{};
    m_device->CreateDepthStencilView(depthTexture.Get(), &depthStencilViewDesc, &depthStencilView);
    m_depthMap.insert(std::make_pair(key, depthStencilView));
    return depthStencilView;
}

ComPtr<ID3D11RenderTargetView> D3D11Renderer::getRenderTarget(ID3D11Texture2D* texture) const
{
    ComPtr<ID3D11RenderTargetView> renderTarget;
    D3D11_RENDER_TARGET_VIEW_DESC rtvDesc{};
    rtvDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
    rtvDesc.ViewDimension = D3D11_RTV_DIMENSION_TEXTURE2D;
    rtvDesc.Texture2D.MipSlice = 0;
    HCHECK(m_device->CreateRenderTargetView(texture, &rtvDesc, renderTarget.GetAddressOf()));
    return renderTarget;
}


void D3D11Renderer::createDepthStencilStates()
{
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc{};
    depthStencilDesc.StencilEnable = false;

    depthStencilDesc.DepthEnable = true;
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;

    HCHECK(m_device->CreateDepthStencilState(&depthStencilDesc, &m_depthStencilState));
}

void D3D11Renderer::createRasterizerState()
{
    D3D11_RASTERIZER_DESC rasterizerDesc{};
    rasterizerDesc.FillMode = D3D11_FILL_SOLID;
    rasterizerDesc.CullMode = D3D11_CULL_BACK;
    rasterizerDesc.FrontCounterClockwise = true;
    rasterizerDesc.DepthBias = 0;
    rasterizerDesc.DepthBiasClamp = 0;
    rasterizerDesc.SlopeScaledDepthBias = 1.0f;
    rasterizerDesc.DepthClipEnable = true;
    rasterizerDesc.ScissorEnable = false;
    rasterizerDesc.MultisampleEnable = false;
    rasterizerDesc.AntialiasedLineEnable = false;

    HCHECK(m_device->CreateRasterizerState(&rasterizerDesc, &m_rasterizerState));
}
