#pragma once

#include <map>
#include <unordered_map>
#include "IRenderer.hpp"

//#define SHOW_WINDOW 1
#ifdef SHOW_WINDOW
#include "Window.h"
#endif

class D3D11Renderer : public IRenderer
{
public:
    D3D11Renderer(XrInstance instance, XrSystemId systemId, bool mrEnabled);

    int64_t getColorTextureFormat() override { return DXGI_FORMAT_R8G8B8A8_UNORM_SRGB; }
    int64_t getDepthMapTextureFormat() override { return DXGI_FORMAT_R24G8_TYPELESS; }
    void insertColorSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) override;
    void insertDepthSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) override;
    void update(float deltaTime) override;
    void render(
        XrSwapchain colorHandle, XrSwapchain depthHandle, uint32_t colorIndex, uint32_t depthIndex, const XrCompositionLayerProjectionView& view) override;
    void* getGraphicsBindings() override { return &m_graphicsBindings; }
    void copyForWindow(XrSwapchain handle, int index) override;
    bool showWindow() override;

private:
    void initCube(float posX, float posY, float posZ, float alpha);
    void initConstants();
    void initShaders();
    static void enumerateSwapchainImages(XrSwapchain handle, std::unordered_map<XrSwapchain, std::vector<XrSwapchainImageD3D11KHR>>& swapchainImagesMap);

    void createBuf(void* bufdata, int datasize, UINT bindFlags, ID3D11Buffer** buf) const;
    static ComPtr<ID3DBlob> compileShader(const char* src, const char* target);
    void clear(ID3D11RenderTargetView* colorRTV, ID3D11DepthStencilView* depthView) const;
    ComPtr<ID3D11DepthStencilView> getDepthMap(XrSwapchain colorSwapchain, XrSwapchain depthSwapchain, uint32_t colorIndex, uint32_t depthIndex);
    ComPtr<ID3D11RenderTargetView> getRenderTarget(ID3D11Texture2D* texture) const;
    ID3D11Device* GetD3DDevice() const { return m_device.Get(); }
    void createDepthStencilStates();
    void createRasterizerState();

    struct Cube {
        float posX;
        float posY;
        float posZ;
        ComPtr<ID3D11Buffer> vertexBuffer;
        int vertexStride;
        ComPtr<ID3D11Buffer> indexBuffer;
        int indexCount;
    };
    std::vector<Cube> m_cubes;
    float m_cubeRot;

    struct Constants {
        struct ConstantBufferStruct {
            DirectX::XMFLOAT4X4 world;
            DirectX::XMFLOAT4X4 view;
            DirectX::XMFLOAT4X4 projection;
        } m_data;
        static_assert((sizeof(ConstantBufferStruct) % 16) == 0, "Constant Buffer size must be 16-byte aligned");
        ComPtr<ID3D11Buffer> m_constantBuffer;
    } m_constants;

    struct Shaders {
        ComPtr<ID3D11VertexShader> vertexShader;
        ComPtr<ID3D11InputLayout> inputLayout;
        ComPtr<ID3D11PixelShader> pixelShader;
    } m_shaders;

    ComPtr<ID3D11Device> m_device;
    ComPtr<ID3D11DeviceContext> m_dc;
    std::unordered_map<XrSwapchain, std::vector<XrSwapchainImageD3D11KHR>> m_colorSwapchainImagesMap;
    std::unordered_map<XrSwapchain, std::vector<XrSwapchainImageD3D11KHR>> m_depthSwapchainImagesMap;
    std::map<std::pair<XrSwapchain, uint32_t>, ComPtr<ID3D11DepthStencilView>> m_depthMap;
    XrGraphicsBindingD3D11KHR m_graphicsBindings{XR_TYPE_GRAPHICS_BINDING_D3D11_KHR};
    ComPtr<ID3D11DepthStencilState> m_depthStencilState;
    ComPtr<ID3D11RasterizerState> m_rasterizerState;
#ifdef SHOW_WINDOW
    std::unique_ptr<Window> m_window;
#endif
};
