#include "stdafx.h"
#include "OpenGLRenderer.hpp"
#include <glm/gtc/type_ptr.hpp>

#define CHECK_GL_ERROR(statement)                                                \
    statement;                                                                   \
    {                                                                            \
        GLenum error = glGetError();                                             \
        if (error != GL_NO_ERROR) {                                              \
            printf("%s: %s: GL error %#06x\n", __FUNCTION__, #statement, error); \
            abort();                                                             \
        }                                                                        \
    }

namespace
{
glm::mat4 poseToViewMatrix(const XrPosef& pose)
{
    auto mat = glm::mat4_cast(glm::dquat(pose.orientation.w, pose.orientation.x, pose.orientation.y, pose.orientation.z));
    mat[3][0] = pose.position.x;
    mat[3][1] = pose.position.y;
    mat[3][2] = pose.position.z;
    return glm::inverse(mat);
}

glm::mat4 fovToProjMatrix(const XrFovf& fov)
{
    const glm::dvec2 min{tan(fov.angleLeft), tan(fov.angleDown)}, max{tan(fov.angleRight), tan(fov.angleUp)};
    const glm::dvec2 invSize{1.0 / (max - min)};
    const glm::dvec2 offset{(min + max) * invSize};
    glm::mat4 projectionMatrix = {
        {2.0 * invSize.x, 0, 0, 0},
        {0, 2.0 * invSize.y, 0, 0},
        {offset.x, offset.y, -1.0, -1.0},
        {0, 0, -0.1, 0},
    };
    return projectionMatrix;
}

void identityMatrix(float* m)
{
    for (int i = 0; i < 16; ++i) m[i] = 0.0f;

    m[0] = 1.0f;
    m[5] = 1.0f;
    m[10] = 1.0f;
    m[15] = 1.0f;
}

void createTransformMatrix(float* m, float xOffset, float yOffset, float zOffset, float radians)
{
    float rSin = sin(radians);
    float rCos = cos(radians);

    identityMatrix(m);
    m[0] = rCos;
    m[2] = -rSin;
    m[8] = rSin;
    m[10] = rCos;
    m[12] = xOffset;
    m[13] = yOffset;
    m[14] = zOffset;
}

void applyToMatrix(glm::mat4& input, float* mappedMatrix)
{
    for (int i = 0; i < 4; ++i) {
        for (int j = 0; j < 4; ++j) {
            mappedMatrix[i * 4 + j] = input[i][j];
        }
    }
}

const char* GetErrorMsg(const unsigned long error)
{
    static CHAR buffer[MAX_PATH];

    // Call FormatMessage to get error string from winapi
    LPWSTR _buffer = nullptr;
    DWORD size = FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, nullptr, error,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPWSTR)&_buffer, 0, nullptr);

    if (size > 0) {
        char* descbuffer = new char[size * 2];
        size_t convertCount;
        wcstombs_s(&convertCount, descbuffer, size * 2, _buffer, size);
        sprintf_s(buffer, "0x%X: %s", error, descbuffer);
        delete[] descbuffer;
    } else {
        sprintf_s(buffer, "0x%X", error);
    }
    LocalFree(_buffer);

    // trim leading newlines, some error messages might contain them
    int i = (int)(strlen(buffer) - 1);
    for (; i >= 0 && (buffer[i] == '\n' || buffer[i] == '\r'); i--) {
        buffer[i] = '\0';
    }

    return buffer;
}

const char* GetLastErrorString()
{
    DWORD err = GetLastError();
    return err ? GetErrorMsg(err) : "";
}
}  // namespace


OpenGLRenderer::OpenGLRenderer(XrInstance instance, XrSystemId systemId, bool mrEnabled)
    : IRenderer(instance, systemId)
{
    PFN_xrGetOpenGLGraphicsRequirementsKHR pfnGetOpenGLGraphicsRequirementsKHR = nullptr;
    XRCHECK(xrGetInstanceProcAddr(instance, "xrGetOpenGLGraphicsRequirementsKHR", reinterpret_cast<PFN_xrVoidFunction*>(&pfnGetOpenGLGraphicsRequirementsKHR)));

    XrGraphicsRequirementsOpenGLKHR glGraphicsRequirements{XR_TYPE_GRAPHICS_REQUIREMENTS_OPENGL_KHR};
    XRCHECK(pfnGetOpenGLGraphicsRequirementsKHR(instance, systemId, &glGraphicsRequirements));

    createWindow();

    if (wglewInit() != GLEW_OK) {
        printf("Failed to initialize WGLEW\n");
        return;
    }

    m_hdc = GetDC(m_hwnd);
    if (m_hdc == NULL) {
        printf("Failed to get DC: %s\n", GetLastErrorString());
        return;
    }

    PIXELFORMATDESCRIPTOR pfd = {sizeof(PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,  // Flags
        PFD_TYPE_RGBA,                                               // The kind of framebuffer. RGBA or palette.
        32,                                                          // Colordepth of the framebuffer.
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        24,  // Number of bits for the depthbuffer
        8,   // Number of bits for the stencilbuffer
        0,   // Number of Aux buffers in the framebuffer.
        PFD_MAIN_PLANE, 0, 0, 0, 0};

    int pixelFormat = ChoosePixelFormat(m_hdc, &pfd);

    if (SetPixelFormat(m_hdc, pixelFormat, &pfd) == false) {
        printf("Failed to set pixel format: %s", GetLastErrorString());
        return;
    }

    m_hglrc = wglCreateContext(m_hdc);
    if (m_hglrc == NULL) {
        printf("Failed to create OpenGL context: %s\n", GetLastErrorString());
        return;
    }

    if (wglMakeCurrent(m_hdc, m_hglrc) == false) {
        printf("Failed to set current OpenGL context: %s\n", GetLastErrorString());
        return;
    }

    if (glewInit() != GLEW_OK) {
        printf("Failed to initialize GLEW\n");
        return;
    }

    glEnable(GL_FRAMEBUFFER_SRGB);

    compileShaders();
    createCube( 0.0f, 0.0f, -1.0f, 1.0f);
    createCube(-1.5f, 0.0f, -3.0f, mrEnabled ? 0.6f : 1.0f);
    m_cubeRot = 0;

    m_graphicsBindings.hDC = m_hdc;
    m_graphicsBindings.hGLRC = m_hglrc;
}

OpenGLRenderer::~OpenGLRenderer()
{
    glDeleteProgram(m_shaderProgram);

    for (const Cube& cube : m_cubes) {
        glDeleteBuffers(1, &cube.vertexBuffer);
        glDeleteBuffers(1, &cube.indexBuffer);
        glDeleteVertexArrays(1, &cube.vao);
    }

    wglMakeCurrent(NULL, NULL);

    if (m_depthBuffer) {
        glDeleteRenderbuffers(1, &m_depthBuffer);
    }

    for (const auto& [swapchain, renderTargets] : m_renderTargetsMap) {
        for (const auto& renderTargetsDepth : renderTargets) {
            for (const auto& renderTarget : renderTargetsDepth) {
                glDeleteFramebuffers(1, &renderTarget.frameBuffer);
            }
        }
    }
}

void OpenGLRenderer::insertColorSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) { m_swapchainSizeMap[handle] = Size{width, height}; }

void OpenGLRenderer::insertDepthSwapchain(XrSwapchain handle, uint32_t width, uint32_t height) {}

bool OpenGLRenderer::initRenderTargets(XrSwapchain color, XrSwapchain depth)
{
    const std::pair<XrSwapchain, XrSwapchain> key = std::make_pair(color, depth);
    const std::vector<XrSwapchainImageOpenGLKHR> colorImages = getImagesFromSwapChain(color);
    std::vector<XrSwapchainImageOpenGLKHR> depthImages;
    if (depth != XR_NULL_HANDLE) {
        depthImages = getImagesFromSwapChain(depth);
    } else {
        glGenRenderbuffers(1, &m_depthBuffer);
        glBindRenderbuffer(GL_RENDERBUFFER, m_depthBuffer);
        const Size size = m_swapchainSizeMap[color];
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, size.width, size.height);
    }

    m_renderTargetsMap[key].resize(colorImages.size());
    for (size_t i = 0; i < colorImages.size(); i++) {
        const size_t depthImagesSize = depth == XR_NULL_HANDLE ? 1 : depthImages.size();
        std::vector<RenderTarget>& renderTargetsDepth = m_renderTargetsMap[key][i];
        renderTargetsDepth.resize(depthImagesSize);
        for (size_t j = 0; j < depthImagesSize; j++) {
            RenderTarget& renderTarget = renderTargetsDepth[j];

            glGenFramebuffers(1, &renderTarget.frameBuffer);
            glBindFramebuffer(GL_FRAMEBUFFER, renderTarget.frameBuffer);

            glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, colorImages[i].image, 0);
            if (depth != XR_NULL_HANDLE) {
                glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, depthImages[j].image, 0);
            } else {
                glBindRenderbuffer(GL_RENDERBUFFER, m_depthBuffer);
                glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, m_depthBuffer);
            }

            GLenum drawBuffers[1] = {GL_COLOR_ATTACHMENT0};
            glDrawBuffers(1, drawBuffers);

            const GLenum error = glGetError();
            if (error != GL_NO_ERROR) {
                printf("Failed to create render buffer: %x", error);
                return false;
            }

            const GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
            if (status != GL_FRAMEBUFFER_COMPLETE) {
                printf("Incomplete framebuffer: %x\n", status);
                return false;
            }
        }
    }

    return true;
}

OpenGLRenderer::RenderTarget OpenGLRenderer::getRenderTarget(XrSwapchain color, XrSwapchain depth, uint32_t colorIndex, uint32_t depthIndex)
{
    const std::vector<std::vector<RenderTarget>>& renderTargets = m_renderTargetsMap[std::make_pair(color, depth)];
    if (renderTargets.empty()) {
        initRenderTargets(color, depth);
    }

    return renderTargets[colorIndex][depth == XR_NULL_HANDLE ? 0 : depthIndex];
}

std::vector<XrSwapchainImageOpenGLKHR> OpenGLRenderer::getImagesFromSwapChain(XrSwapchain swapchain)
{
    uint32_t imageCount;
    XRCHECK(xrEnumerateSwapchainImages(swapchain, 0, &imageCount, nullptr));

    std::vector<XrSwapchainImageOpenGLKHR> glSwapchainImages(imageCount);
    for (uint32_t i = 0; i < imageCount; i++) {
        glSwapchainImages[i] = XrSwapchainImageOpenGLKHR{XR_TYPE_SWAPCHAIN_IMAGE_OPENGL_KHR};
    }

    XRCHECK(xrEnumerateSwapchainImages(swapchain, imageCount, &imageCount, reinterpret_cast<XrSwapchainImageBaseHeader*>(glSwapchainImages.data())));
    return glSwapchainImages;
}

void OpenGLRenderer::update(float deltaTime) { m_cubeRot += 0.35f * deltaTime; }

void OpenGLRenderer::render(XrSwapchain color, XrSwapchain depth, uint32_t colorIndex, uint32_t depthIndex, const XrCompositionLayerProjectionView& view)
{
    const RenderTarget renderTarget = getRenderTarget(color, depth, colorIndex, depthIndex);
    CHECK_GL_ERROR(glBindFramebuffer(GL_FRAMEBUFFER, renderTarget.frameBuffer));
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    CHECK_GL_ERROR(glClearColor(0.0f, 0.0f, 0.0f, 0.0f));
    CHECK_GL_ERROR(glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));

    CHECK_GL_ERROR(glViewport(
        view.subImage.imageRect.offset.x, view.subImage.imageRect.offset.y, view.subImage.imageRect.extent.width, view.subImage.imageRect.extent.height));

    glUseProgram(m_shaderProgram);

    for (const Cube& cube : m_cubes) {
        glBindVertexArray(cube.vao);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cube.indexBuffer);

        CHECK_GL_ERROR(glBindBuffer(GL_UNIFORM_BUFFER, m_uniformBuffer.buffer));
        CHECK_GL_ERROR(float* mappedUniforms = (float*)glMapBufferRange(GL_UNIFORM_BUFFER, 0, sizeof(m_uniformBuffer.data), GL_MAP_WRITE_BIT));

        auto viewMatrix = poseToViewMatrix(view.pose);
        auto projectionMatrix = fovToProjMatrix(view.fov);
        createTransformMatrix(&mappedUniforms[0], cube.posX, cube.posY, cube.posZ, m_cubeRot);
        applyToMatrix(viewMatrix, &mappedUniforms[16]);
        applyToMatrix(projectionMatrix, &mappedUniforms[32]);

        CHECK_GL_ERROR(glUnmapBuffer(GL_UNIFORM_BUFFER));
        CHECK_GL_ERROR(glBindBufferBase(GL_UNIFORM_BUFFER, 0, m_uniformBuffer.buffer));
        CHECK_GL_ERROR(glUniformBlockBinding(m_shaderProgram, 0, 0));

        CHECK_GL_ERROR(glDrawElements(GL_TRIANGLES, cube.indexCount, GL_UNSIGNED_SHORT, 0));
    }

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void OpenGLRenderer::compileShaders()
{
    GLuint vertexShader = compileShader(GL_VERTEX_SHADER,
        "#version 430\n"
        "layout(location = 0) in vec3 position;\n"
        "layout(location = 1) in vec4 color;\n"
        "layout(std140, binding = 0) uniform Matrices {\n"
        "	mat4 worldMatrix;\n"
        "	mat4 viewMatrix;\n"
        "	mat4 projectionMatrix;\n"
        "};\n"
        "out vec4 vColor;\n"
        "void main()\n"
        "{\n"
        "	vColor = vec4(color.rgb * color.a, color.a);\n"
        "	gl_Position = projectionMatrix * viewMatrix * worldMatrix * vec4(position, 1);\n"
        "}\n");

    GLuint fragmentShader = compileShader(GL_FRAGMENT_SHADER,
        "#version 430\n"
        "in vec4 vColor;\n"
        "out vec4 oColor;\n"
        "void main()\n"
        "{\n"
        "	oColor = vColor;\n"
        "}\n");

    m_shaderProgram = glCreateProgram();
    glAttachShader(m_shaderProgram, vertexShader);
    glAttachShader(m_shaderProgram, fragmentShader);
    glLinkProgram(m_shaderProgram);

    GLint status = 0;
    glGetProgramiv(m_shaderProgram, GL_LINK_STATUS, &status);
    if (status != GL_TRUE) {
        GLint infoSize = 0;
        glGetProgramiv(m_shaderProgram, GL_INFO_LOG_LENGTH, &infoSize);

        std::vector<char> infoLog(infoSize);
        glGetProgramInfoLog(m_shaderProgram, infoSize, &infoSize, infoLog.data());

        printf("Failed to link program:\n");
        printf("%s\n", infoLog.data());
        abort();
    }
}

GLuint OpenGLRenderer::compileShader(GLenum type, const char* source)
{
    GLuint shader = glCreateShader(type);
    glShaderSource(shader, 1, &source, nullptr);
    glCompileShader(shader);

    GLint status = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        GLint infoSize = 0;
        glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoSize);

        std::vector<char> infoLog(infoSize);
        glGetShaderInfoLog(shader, infoSize, &infoSize, infoLog.data());

        printf("Failed to compile shader:\n");
        printf("%s\n", infoLog.data());
        abort();
    }

    return shader;
}

void OpenGLRenderer::createCube(float posX, float posY, float posZ, float alpha)
{
    m_cubes.emplace_back();
    Cube& cube = m_cubes.back();

    cube.posX = posX;
    cube.posY = posY;
    cube.posZ = posZ;

    float vertices[][7]{
        {-0.1f, -0.1f, -0.1f, 0, 0, 0, alpha},
        {-0.1f, -0.1f,  0.1f, 0, 0, 1, alpha},
        {-0.1f,  0.1f, -0.1f, 0, 1, 0, alpha},
        {-0.1f,  0.1f,  0.1f, 0, 1, 1, alpha},
        { 0.1f, -0.1f, -0.1f, 1, 0, 0, alpha},
        { 0.1f, -0.1f,  0.1f, 1, 0, 1, alpha},
        { 0.1f,  0.1f, -0.1f, 1, 1, 0, alpha},
        { 0.1f,  0.1f,  0.1f, 1, 1, 1, alpha},
    };

    glGenVertexArrays(1, &cube.vao);
    glBindVertexArray(cube.vao);

    glGenBuffers(1, &cube.vertexBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, cube.vertexBuffer);
    CHECK_GL_ERROR(glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW));

    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, false, sizeof(float) * 7, 0);

    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 4, GL_FLOAT, false, sizeof(float) * 7, (void*)(sizeof(float) * 3));

    unsigned short indices[][3]{
        {0, 2, 1},
        {1, 2, 3},
        {4, 5, 6},
        {5, 7, 6},
        {0, 1, 5},
        {0, 5, 4},
        {2, 6, 7},
        {2, 7, 3},
        {0, 4, 6},
        {0, 6, 2},
        {1, 3, 7},
        {1, 7, 5},
    };


    cube.indexCount = sizeof indices / sizeof(unsigned short);

    glGenBuffers(1, &cube.indexBuffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cube.indexBuffer);
    CHECK_GL_ERROR(glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW));

    glBindVertexArray(0);

    glGenBuffers(1, &m_uniformBuffer.buffer);
    glBindBuffer(GL_UNIFORM_BUFFER, m_uniformBuffer.buffer);
    CHECK_GL_ERROR(glBufferData(GL_UNIFORM_BUFFER, sizeof(m_uniformBuffer.data), nullptr, GL_STREAM_DRAW));
    glBindBuffer(GL_UNIFORM_BUFFER, 0);
}

void OpenGLRenderer::createWindow()
{
    const int WIDTH = 800;
    const int HEIGHT = 600;

    static char szWindowClass[] = "Varjo";
    static char szTitle[] = "Varjo";

    WNDCLASSEXA wcex;

    HINSTANCE hInstance = GetModuleHandle(nullptr);

    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = wndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = LoadIcon(hInstance, NULL);
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wcex.lpszMenuName = NULL;
    wcex.lpszClassName = szWindowClass;
    wcex.hIconSm = LoadIcon(wcex.hInstance, NULL);

    if (!RegisterClassExA(&wcex)) {
        printf("RegisterClassEx failed: %s\n", GetLastErrorString());
        return;
    }

    m_hwnd = CreateWindowA(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, WIDTH, HEIGHT, NULL, NULL, hInstance, NULL);
    if (!m_hwnd) {
        printf("CreateWindow failed: %s\n", GetLastErrorString());
        return;
    }

    UpdateWindow(m_hwnd);
}