#include "stdafx.h"
#include <map>
#include <unordered_map>
#include <unordered_set>
#include <iostream>
#include "OpenGLRenderer.hpp"
#include "D3D11Renderer.hpp"
#include "Window.h"

#pragma comment(lib, "d3dcompiler.lib")
#pragma comment(lib, "d3d11.lib")
#pragma comment(lib, "dxgi.lib")

using namespace std;

struct SwapChain {
    XrSwapchain handle;
    uint32_t width;
    uint32_t height;
};

bool gotKey()
{
    HANDLE in = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD input;
    DWORD count = 0;
    while (GetNumberOfConsoleInputEvents(in, &count) && count > 0) {
        if (ReadConsoleInputA(in, &input, 1, &count) && count > 0 && input.EventType == KEY_EVENT && input.Event.KeyEvent.bKeyDown) {
            if (input.Event.KeyEvent.uChar.AsciiChar == '\033') {
                printf("Quit requested.\n");
                return true;
            }
        }
    }
    return false;
}

#pragma region State Management
bool isVisible(XrSessionState sessionState) { return sessionState == XR_SESSION_STATE_FOCUSED || sessionState == XR_SESSION_STATE_VISIBLE; }

#pragma endregion

#define LOAD_XR_FUNCTION(instance, function)                                                                   \
    do {                                                                                                       \
        static_assert(std::is_same<decltype(function), PFN_##function>::value);                                \
        XRCHECK(xrGetInstanceProcAddr(instance, #function, reinterpret_cast<PFN_xrVoidFunction*>(&function))); \
    } while (false)

bool hasExtension(const char* name)
{
    uint32_t extensionsCount;
    XRCHECK(xrEnumerateInstanceExtensionProperties(nullptr, 0, &extensionsCount, nullptr));
    std::vector<XrExtensionProperties> properties(extensionsCount);
    for (auto& prop : properties) {
        prop.type = XR_TYPE_EXTENSION_PROPERTIES;
    }
    XRCHECK(xrEnumerateInstanceExtensionProperties(nullptr, extensionsCount, &extensionsCount, properties.data()));

    for (const auto& prop : properties) {
        if (strcmp(prop.extensionName, name) == 0) {
            return true;
        }
    }
    return false;
}

bool hasArgument(int argc, char** argv, const char* parameter)
{
    for (int i = 1; i < argc; ++i) {
        if (strcmp(argv[i], parameter) == 0) {
            return true;
        }
    }
    return false;
}

int main(int argc, char** argv)
{
    const bool useOpengl = hasArgument(argc, argv, "--use-opengl");
    const bool stillScene = hasArgument(argc, argv, "--still-scene");
    const bool mrEnabled = hasArgument(argc, argv, "--use-mr");
    const bool useDepthTest = hasArgument(argc, argv, "--use-depth-test");
    const bool useDepthTestRange = hasArgument(argc, argv, "--use-depth-test-range");
    const bool useEnvironmentDepthEstimation = hasArgument(argc, argv, "--use-environment-depth-estimation");

#pragma region Boilerplate
    uint32_t instanceExtensionCount;
    XRCHECK(xrEnumerateInstanceExtensionProperties(nullptr, 0, &instanceExtensionCount, nullptr));

    std::vector<XrExtensionProperties> extensions(instanceExtensionCount);
    for (XrExtensionProperties& extension : extensions) {
        extension.type = XR_TYPE_EXTENSION_PROPERTIES;
    }
    XRCHECK(xrEnumerateInstanceExtensionProperties(nullptr, static_cast<uint32_t>(extensions.size()), &instanceExtensionCount, extensions.data()));
    std::cout << "Supported extensions:\n";
    for (const auto& ext : extensions) {
        std::cout << "    " << ext.extensionName << ", " << ext.extensionVersion << std::endl;
    }

    //  xrCreateInstance
    std::vector<const char*> requiredExtensions{XR_KHR_D3D11_ENABLE_EXTENSION_NAME};
    if (hasExtension(XR_KHR_OPENGL_ENABLE_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_KHR_OPENGL_ENABLE_EXTENSION_NAME);
    }

    if (hasExtension(XR_VARJO_QUAD_VIEWS_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_VARJO_QUAD_VIEWS_EXTENSION_NAME);
    }

    bool performanceCounterTimeExtensionEnabled = false;
    if (hasExtension(XR_KHR_WIN32_CONVERT_PERFORMANCE_COUNTER_TIME_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_KHR_WIN32_CONVERT_PERFORMANCE_COUNTER_TIME_EXTENSION_NAME);
        performanceCounterTimeExtensionEnabled = true;
    }

    bool depthExtEnabled = false;
    if (hasExtension(XR_KHR_COMPOSITION_LAYER_DEPTH_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_KHR_COMPOSITION_LAYER_DEPTH_EXTENSION_NAME);
        depthExtEnabled = true;
    }

    bool gazeInteractionEnabled = false;
    if (hasExtension(XR_EXT_EYE_GAZE_INTERACTION_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_EXT_EYE_GAZE_INTERACTION_EXTENSION_NAME);
        gazeInteractionEnabled = true;
    }

    bool foveationExtensionEnabled = false;
    if (hasExtension(XR_VARJO_FOVEATED_RENDERING_EXTENSION_NAME)) {
        requiredExtensions.push_back(XR_VARJO_FOVEATED_RENDERING_EXTENSION_NAME);
        foveationExtensionEnabled = true;
    }

    bool depthTestEnabled = false;
    if (useDepthTest) {
        if (hasExtension(XR_VARJO_COMPOSITION_LAYER_DEPTH_TEST_EXTENSION_NAME)) {
            requiredExtensions.push_back(XR_VARJO_COMPOSITION_LAYER_DEPTH_TEST_EXTENSION_NAME);
            depthTestEnabled = true;
        } else {
            printf("Cannot enable depth test - " XR_VARJO_COMPOSITION_LAYER_DEPTH_TEST_EXTENSION_NAME " not supported\n");
            abort();
        }
    }

    bool depthTestRangeEnabled = false;
    if (useDepthTestRange) {
        if (depthTestEnabled) {
            depthTestRangeEnabled = true;
        } else {
            printf("Please enable depth test to use depth test range\n");
            abort();
        }
    }

    bool environmentDepthEstimationEnabled = false;
    if (useEnvironmentDepthEstimation) {
        if (!mrEnabled) {
            printf("Please enable mr to use environment depth estimation\n");
            abort();
        } else if (!depthExtEnabled) {
            printf("Cannot enable environment depth estimation - " XR_KHR_COMPOSITION_LAYER_DEPTH_EXTENSION_NAME " extension not supported\n");
            abort();
        } else if (!hasExtension(XR_VARJO_ENVIRONMENT_DEPTH_ESTIMATION_EXTENSION_NAME)) {
            printf("Cannot enable environment depth estimation - " XR_VARJO_ENVIRONMENT_DEPTH_ESTIMATION_EXTENSION_NAME " extension not supported\n");
            abort();
        } else {
            requiredExtensions.push_back(XR_VARJO_ENVIRONMENT_DEPTH_ESTIMATION_EXTENSION_NAME);
            environmentDepthEstimationEnabled = true;
        }
    }

    using namespace std;

    XrInstanceCreateInfo createInfo{XR_TYPE_INSTANCE_CREATE_INFO};
    createInfo.next = nullptr;
    createInfo.enabledExtensionCount = static_cast<uint32_t>(requiredExtensions.size());
    createInfo.enabledExtensionNames = requiredExtensions.data();

    strcpy_s(createInfo.applicationInfo.applicationName, "HelloXR");
    createInfo.applicationInfo.apiVersion = XR_CURRENT_API_VERSION;

    XrInstance instance;
    XRCHECK(xrCreateInstance(&createInfo, &instance));

    PFN_xrSetEnvironmentDepthEstimationVARJO xrSetEnvironmentDepthEstimationVARJO = nullptr;
    if (environmentDepthEstimationEnabled) {
        LOAD_XR_FUNCTION(instance, xrSetEnvironmentDepthEstimationVARJO);
    }

    //  xrGetSystem
    XrSystemGetInfo systemInfo{XR_TYPE_SYSTEM_GET_INFO};
    systemInfo.formFactor = XR_FORM_FACTOR_HEAD_MOUNTED_DISPLAY;
    XrSystemId systemId;
    XRCHECK(xrGetSystem(instance, &systemInfo, &systemId));

    // xrEnumerateViewConfigurations
    uint32_t viewConfigTypeCount;
    XRCHECK(xrEnumerateViewConfigurations(instance, systemId, 0, &viewConfigTypeCount, nullptr));
    std::vector<XrViewConfigurationType> viewConfigTypes(viewConfigTypeCount);
    XRCHECK(xrEnumerateViewConfigurations(instance, systemId, viewConfigTypeCount, &viewConfigTypeCount, viewConfigTypes.data()));

    if (viewConfigTypes.empty()) {
        printf("Don't have good configs\n");
        abort();
    }

    XrViewConfigurationType viewConfigType = viewConfigTypes[0];  // Pick first view config

    // xrEnumerateEnvironmentBlendModes
    uint32_t environmentBlendModeCount;
    XRCHECK(xrEnumerateEnvironmentBlendModes(instance, systemId, viewConfigType, 0, &environmentBlendModeCount, nullptr));
    std::vector<XrEnvironmentBlendMode> environmentBlendModes(environmentBlendModeCount);
    XRCHECK(xrEnumerateEnvironmentBlendModes(
        instance, systemId, viewConfigType, environmentBlendModeCount, &environmentBlendModeCount, environmentBlendModes.data()));

    XrEnvironmentBlendMode environmentBlendMode = mrEnabled ? XR_ENVIRONMENT_BLEND_MODE_ALPHA_BLEND : XR_ENVIRONMENT_BLEND_MODE_OPAQUE;
    if (std::find(environmentBlendModes.cbegin(), environmentBlendModes.cend(), environmentBlendMode) == environmentBlendModes.cend()) {
        printf("Environment blend mode appropriate for selected mr configuration is not supported\n");
        abort();
    }

    // xrGetSystemProperties
    XrSystemProperties systemProperties{XR_TYPE_SYSTEM_PROPERTIES};
    XrSystemFoveatedRenderingPropertiesVARJO foveatedRenderingProperties{XR_TYPE_SYSTEM_FOVEATED_RENDERING_PROPERTIES_VARJO};
    if (foveationExtensionEnabled) {
        systemProperties.next = &foveatedRenderingProperties;
    }

    XrSystemEyeGazeInteractionPropertiesEXT gazeInteractionProperties{XR_TYPE_SYSTEM_EYE_GAZE_INTERACTION_PROPERTIES_EXT};
    if (gazeInteractionEnabled) {
        gazeInteractionProperties.next = systemProperties.next;
        systemProperties.next = &gazeInteractionProperties;
    }
    XRCHECK(xrGetSystemProperties(instance, systemId, &systemProperties));

    uint32_t viewCount;
    XRCHECK(xrEnumerateViewConfigurationViews(instance, systemId, viewConfigType, 0, &viewCount, nullptr));
    std::vector<XrViewConfigurationView> configViews(viewCount, {XR_TYPE_VIEW_CONFIGURATION_VIEW});
    XRCHECK(xrEnumerateViewConfigurationViews(instance, systemId, viewConfigType, viewCount, &viewCount, configViews.data()));

    const bool foveationSupported =
        foveationExtensionEnabled && foveatedRenderingProperties.supportsFoveatedRendering && viewConfigType == XR_VIEW_CONFIGURATION_TYPE_PRIMARY_QUAD_VARJO;
    std::vector<XrViewConfigurationView> foveatedViews;
    if (foveationSupported) {
        std::vector<XrFoveatedViewConfigurationViewVARJO> requestFoveatedConfig{viewCount, {XR_TYPE_FOVEATED_VIEW_CONFIGURATION_VIEW_VARJO, nullptr, XR_TRUE}};
        foveatedViews = std::vector<XrViewConfigurationView>(viewCount, {XR_TYPE_VIEW_CONFIGURATION_VIEW});
        for (uint32_t i = 0; i < viewCount; i++) {
            foveatedViews[i].next = &requestFoveatedConfig[i];
        }
        XRCHECK(xrEnumerateViewConfigurationViews(instance, systemId, viewConfigType, viewCount, &viewCount, foveatedViews.data()));
    }

    std::vector<XrView> views;
    views.resize(viewCount);

    std::unique_ptr<IRenderer> renderer;
    if (useOpengl)
        renderer = std::make_unique<OpenGLRenderer>(instance, systemId, mrEnabled);
    else
        renderer = std::make_unique<D3D11Renderer>(instance, systemId, mrEnabled);

    // xrCreateSession
    XrSession session;
    XrSessionCreateInfo sessionCreateInfo{XR_TYPE_SESSION_CREATE_INFO};
    sessionCreateInfo.next = static_cast<const XrBaseInStructure*>(renderer->getGraphicsBindings());
    sessionCreateInfo.systemId = systemId;
    XRCHECK(xrCreateSession(instance, &sessionCreateInfo, &session));

    // xrEnumerateSwapchainFormats
    uint32_t swapchainFormatCount;
    XRCHECK(xrEnumerateSwapchainFormats(session, 0, &swapchainFormatCount, nullptr));
    std::vector<int64_t> swapchainFormats(swapchainFormatCount);
    XRCHECK(xrEnumerateSwapchainFormats(session, static_cast<uint32_t>(swapchainFormats.size()), &swapchainFormatCount, swapchainFormats.data()));

    // TODO select texture format more dynamically
    if (std::find(swapchainFormats.cbegin(), swapchainFormats.cend(), renderer->getColorTextureFormat()) == swapchainFormats.cend()) {
        printf("Can't find %s as a supported texture format", useOpengl ? "GL_SRGB8_ALPHA8" : "DXGI_FORMAT_R8G8B8A8_UNORM_SRGB");
        exit(0);
    }

    ///
    uint32_t spaceCount;
    XRCHECK(xrEnumerateReferenceSpaces(session, 0, &spaceCount, nullptr));
    std::vector<XrReferenceSpaceType> spaces(spaceCount);
    XRCHECK(xrEnumerateReferenceSpaces(session, spaceCount, &spaceCount, spaces.data()));
    ///

    std::vector<SwapChain> swapchains;
    std::vector<SwapChain> depthSwapchains;

    for (uint32_t i = 0; i < viewCount; i++) {
        const XrViewConfigurationView& view = configViews[i];
        const XrViewConfigurationView* foveatedView = foveationSupported ? &foveatedViews[i] : nullptr;
        XrSwapchainCreateInfo swapchainCreateInfo{XR_TYPE_SWAPCHAIN_CREATE_INFO};
        swapchainCreateInfo.arraySize = 1;
        swapchainCreateInfo.format = renderer->getColorTextureFormat();
        swapchainCreateInfo.width = std::max(view.recommendedImageRectWidth, foveationSupported ? foveatedView->recommendedImageRectWidth : 0);
        swapchainCreateInfo.height = std::max(view.recommendedImageRectHeight, foveationSupported ? foveatedView->recommendedImageRectHeight : 0);
        swapchainCreateInfo.mipCount = 1;
        swapchainCreateInfo.faceCount = 1;
        swapchainCreateInfo.sampleCount = view.recommendedSwapchainSampleCount;
        swapchainCreateInfo.usageFlags = XR_SWAPCHAIN_USAGE_SAMPLED_BIT | XR_SWAPCHAIN_USAGE_COLOR_ATTACHMENT_BIT;

        SwapChain swapchain{nullptr, swapchainCreateInfo.width, swapchainCreateInfo.height};
        XRCHECK(xrCreateSwapchain(session, &swapchainCreateInfo, &swapchain.handle));

        swapchains.push_back(swapchain);

        renderer->insertColorSwapchain(swapchain.handle, swapchain.width, swapchain.height);

        if (depthExtEnabled) {
            XrSwapchainCreateInfo depthSwapChainCreateInfo{XR_TYPE_SWAPCHAIN_CREATE_INFO};
            depthSwapChainCreateInfo.arraySize = 1;
            depthSwapChainCreateInfo.format = renderer->getDepthMapTextureFormat();
            depthSwapChainCreateInfo.width = std::max(view.recommendedImageRectWidth, foveationSupported ? foveatedView->recommendedImageRectWidth : 0);
            depthSwapChainCreateInfo.height = std::max(view.recommendedImageRectHeight, foveationSupported ? foveatedView->recommendedImageRectHeight : 0);
            depthSwapChainCreateInfo.mipCount = 1;
            depthSwapChainCreateInfo.faceCount = 1;
            depthSwapChainCreateInfo.sampleCount = view.recommendedSwapchainSampleCount;
            depthSwapChainCreateInfo.usageFlags = XR_SWAPCHAIN_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT;
            SwapChain depthSwapchain{nullptr, depthSwapChainCreateInfo.width, depthSwapChainCreateInfo.height};
            XRCHECK(xrCreateSwapchain(session, &depthSwapChainCreateInfo, &depthSwapchain.handle));
            depthSwapchains.push_back(depthSwapchain);
            renderer->insertDepthSwapchain(depthSwapchain.handle, depthSwapchain.width, depthSwapchain.height);
        }
    }

    XrPosef identity{};
    identity.orientation.w = 1;
    XrReferenceSpaceCreateInfo referenceSpaceCreateInfo{XR_TYPE_REFERENCE_SPACE_CREATE_INFO, nullptr, XR_REFERENCE_SPACE_TYPE_LOCAL, identity};

    XrSpace appSpace;
    XRCHECK(xrCreateReferenceSpace(session, &referenceSpaceCreateInfo, &appSpace));

    XrSessionState sessionState = XR_SESSION_STATE_VISIBLE;
#pragma endregion

    XrActionSetCreateInfo actionSetInfo{XR_TYPE_ACTION_SET_CREATE_INFO};
    strcpy_s(actionSetInfo.actionSetName, "gameplay");
    strcpy_s(actionSetInfo.localizedActionSetName, "Gameplay");
    actionSetInfo.priority = 0;
    XrActionSet gameplayActionSet;
    XRCHECK(xrCreateActionSet(instance, &actionSetInfo, &gameplayActionSet));

    XrPath leftSubactionPath, rightSubactionPath;
    XRCHECK(xrStringToPath(instance, "/user/hand/left", &leftSubactionPath));
    XRCHECK(xrStringToPath(instance, "/user/hand/right", &rightSubactionPath));
    std::array<XrPath, 2> handSubactionPaths{leftSubactionPath, rightSubactionPath};

    // Create an input action for gripping objects with the left and right hands.
    XrActionCreateInfo actionInfo{XR_TYPE_ACTION_CREATE_INFO};

    // action_type_vector1f
    actionInfo.actionType = XR_ACTION_TYPE_FLOAT_INPUT;
    strcpy_s(actionInfo.actionName, "grip_object");
    strcpy_s(actionInfo.localizedActionName, "Grip Object");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction gripAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &gripAction));

    strcpy_s(actionInfo.actionName, "trigger_value");
    strcpy_s(actionInfo.localizedActionName, "trigger_value");
    XrAction triggerValueAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &triggerValueAction));

    strcpy_s(actionInfo.actionName, "joystick_x");
    strcpy_s(actionInfo.localizedActionName, "joystick_x");
    XrAction thumbstickXAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &thumbstickXAction));

    strcpy_s(actionInfo.actionName, "joystick_y");
    strcpy_s(actionInfo.localizedActionName, "joystick_y");
    XrAction thumbstickYAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &thumbstickYAction));


    // action_type_vector2f
    actionInfo.actionType = XR_ACTION_TYPE_VECTOR2F_INPUT;
    strcpy_s(actionInfo.actionName, "joystick");
    strcpy_s(actionInfo.localizedActionName, "joystick");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction thumbstickAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &thumbstickAction));

    // action_type_boolean
    actionInfo.actionType = XR_ACTION_TYPE_BOOLEAN_INPUT;
    strcpy_s(actionInfo.actionName, "select");
    strcpy_s(actionInfo.localizedActionName, "Select");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction aAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &aAction));

    strcpy_s(actionInfo.actionName, "deselect");
    strcpy_s(actionInfo.localizedActionName, "Deselect");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction bAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &bAction));

    strcpy_s(actionInfo.actionName, "track_touch");
    strcpy_s(actionInfo.localizedActionName, "track_touch");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction trackTouchAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &trackTouchAction));

    strcpy_s(actionInfo.actionName, "track_click");
    strcpy_s(actionInfo.localizedActionName, "track_click");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction trackClickAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &trackClickAction));

    strcpy_s(actionInfo.actionName, "trigger");
    strcpy_s(actionInfo.localizedActionName, "trigger");
    XrAction triggerAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &triggerAction));

    // action_type_pose
    actionInfo.actionType = XR_ACTION_TYPE_POSE_INPUT;
    strcpy_s(actionInfo.actionName, "pose");
    strcpy_s(actionInfo.localizedActionName, "pose");
    actionInfo.countSubactionPaths = static_cast<uint32_t>(handSubactionPaths.size());
    actionInfo.subactionPaths = handSubactionPaths.data();
    XrAction controllersAction;
    XRCHECK(xrCreateAction(gameplayActionSet, &actionInfo, &controllersAction));

    XrSpace leftControllerSpace, rightControllerSpace;
    XrActionSpaceCreateInfo actionSpace{XR_TYPE_ACTION_SPACE_CREATE_INFO};
    actionSpace.action = controllersAction;
    actionSpace.poseInActionSpace.orientation.w = 1.0f;
    actionSpace.subactionPath = leftSubactionPath;
    XRCHECK(xrCreateActionSpace(session, &actionSpace, &leftControllerSpace));

    actionSpace.subactionPath = rightSubactionPath;
    XRCHECK(xrCreateActionSpace(session, &actionSpace, &rightControllerSpace));

    XrPath aButtonClickLeft, aButtonClickRight;
    XrPath bButtonClickLeft, bButtonClickRight;
    XrPath palmPoseLeft, palmPoseRight;
    XrPath gripValueLeft, gripValueRight;
    XrPath gripForceLeft, gripForceRight;
    XrPath triggerClickLeft, triggerClickRight;
    XrPath triggerValueLeft, triggerValueRight;
    XrPath thumbstickXLeft, thumbstickXRight;
    XrPath thumbstickYLeft, thumbstickYRight;
    XrPath thumbstickLeft, thumbstickRight;
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/a/click", &aButtonClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/a/click", &aButtonClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/b/click", &bButtonClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/b/click", &bButtonClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trigger/click", &triggerClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trigger/click", &triggerClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trigger/value", &triggerValueLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trigger/value", &triggerValueRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/palm/pose", &palmPoseLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/palm/pose", &palmPoseRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/grip/value", &gripValueLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/grip/value", &gripValueRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/grip/force", &gripForceLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/grip/force", &gripForceRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/thumbstick/x", &thumbstickXLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/thumbstick/x", &thumbstickXRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/thumbstick/y", &thumbstickYLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/thumbstick/y", &thumbstickYRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/thumbstick", &thumbstickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/thumbstick", &thumbstickRight));

    XrPath gripClickLeft, gripClickRight;
    XrPath menuClickLeft, menuClickRight;
    XrPath triggerClickLeftVive, triggerClickRightVive;
    XrPath trackpadClickLeft, trackpadClickRight;
    XrPath trackpadTouchLeft, trackpadTouchRight;
    XrPath triggerValueLeftVive, triggerValueRightVive;
    XrPath trackpadXLeft, trackpadXRight;
    XrPath trackpadYLeft, trackpadYRight;
    XrPath palmPoseLeftVive, palmPoseRightVive;
    XrPath trackpadLeft, trackpadRight;
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/grip/click", &gripClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/grip/click", &gripClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/menu/click", &menuClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/menu/click", &menuClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trigger/click", &triggerClickLeftVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trigger/click", &triggerClickRightVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trackpad/click", &trackpadClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trackpad/click", &trackpadClickRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trackpad/touch", &trackpadTouchLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trackpad/touch", &trackpadTouchRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trigger/value", &triggerValueLeftVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trigger/value", &triggerValueRightVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trackpad/x", &trackpadXLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trackpad/x", &trackpadXRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trackpad/y", &trackpadYLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trackpad/y", &trackpadYRight));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/palm/pose", &palmPoseLeftVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/palm/pose", &palmPoseRightVive));
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/trackpad", &trackpadLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/trackpad", &trackpadRight));

    XrPath selectClickLeft, selectClickRight;
    XRCHECK(xrStringToPath(instance, "/user/hand/left/input/select/click", &selectClickLeft));
    XRCHECK(xrStringToPath(instance, "/user/hand/right/input/select/click", &selectClickRight));

    XrPath viveKnucklesInteractionProfilePath;
    XRCHECK(xrStringToPath(instance, "/interaction_profiles/valve/knuckles_controller", &viveKnucklesInteractionProfilePath));

    XrPath viveControllerInteractionProfilePath;
    XRCHECK(xrStringToPath(instance, "/interaction_profiles/htc/vive_controller", &viveControllerInteractionProfilePath));

    XrPath simpleControllerInteractionProfilePath;
    XRCHECK(xrStringToPath(instance, "/interaction_profiles/khr/simple_controller", &simpleControllerInteractionProfilePath));

    // clang-format off
    std::array<XrActionSuggestedBinding, 18> knucklesBindings{{
        {aAction, aButtonClickLeft},
        {aAction, aButtonClickRight},
        {bAction, bButtonClickLeft},
        {bAction, bButtonClickRight},
        {controllersAction, palmPoseLeft},
        {controllersAction, palmPoseRight},
        {gripAction, gripValueLeft},
        {gripAction, gripValueRight},
        {triggerAction, triggerClickLeft},
        {triggerAction, triggerClickRight},
        {triggerValueAction, triggerValueLeft},
        {triggerValueAction, triggerValueRight},
        {thumbstickXAction, thumbstickXLeft},
        {thumbstickXAction, thumbstickXRight},
        {thumbstickYAction, thumbstickYLeft},
        {thumbstickYAction, thumbstickYRight},
        {thumbstickAction, thumbstickRight},
        {thumbstickAction, thumbstickLeft},
    }};

    std::array<XrActionSuggestedBinding, 20> viveControllerBindings{{
        {aAction, gripClickLeft},
        {aAction, gripClickRight},
        {bAction, menuClickLeft},
        {bAction, menuClickRight},
        {triggerAction, triggerClickLeftVive},
        {triggerAction, triggerClickRightVive},
        {trackClickAction, trackpadClickLeft},
        {trackClickAction, trackpadClickRight},
        {trackTouchAction, trackpadTouchLeft},
        {trackTouchAction, trackpadTouchRight},
        {triggerValueAction, triggerValueLeftVive},
        {triggerValueAction, triggerValueRightVive},
        {thumbstickXAction, trackpadXLeft},
        {thumbstickXAction, trackpadXRight},
        {thumbstickYAction, trackpadYLeft},
        {thumbstickYAction, trackpadYRight},
        {controllersAction, palmPoseLeftVive},
        {controllersAction, palmPoseRightVive},
        {thumbstickAction, trackpadLeft},
        {thumbstickAction, trackpadRight},
    }};

    std::array<XrActionSuggestedBinding, 4> simpleControllerBindings {{
            {aAction, selectClickLeft},
            {aAction, selectClickRight},
            {bAction, menuClickLeft},
            {bAction, menuClickRight},
    }};

    // clang-format on
    XrInteractionProfileSuggestedBinding suggestedBindings{XR_TYPE_INTERACTION_PROFILE_SUGGESTED_BINDING};
    suggestedBindings.interactionProfile = viveKnucklesInteractionProfilePath;
    suggestedBindings.suggestedBindings = &knucklesBindings[0];
    suggestedBindings.countSuggestedBindings = static_cast<uint32_t>(knucklesBindings.size());
    XrResult res = xrSuggestInteractionProfileBindings(instance, &suggestedBindings);

    suggestedBindings.interactionProfile = viveControllerInteractionProfilePath;
    suggestedBindings.suggestedBindings = &viveControllerBindings[0];
    suggestedBindings.countSuggestedBindings = static_cast<uint32_t>(viveControllerBindings.size());
    res = xrSuggestInteractionProfileBindings(instance, &suggestedBindings);

    suggestedBindings.interactionProfile = simpleControllerInteractionProfilePath;
    suggestedBindings.suggestedBindings = &simpleControllerBindings[0];
    suggestedBindings.countSuggestedBindings = static_cast<uint32_t>(simpleControllerBindings.size());
    res = xrSuggestInteractionProfileBindings(instance, &suggestedBindings);

    XrAction userIntentAction{};
    XrSpace gazeActionSpace{};
    XrSpace localReferenceSpace{};
    vector<XrActionSuggestedBinding> gazeBindings;

    if (gazeInteractionProperties.supportsEyeGazeInteraction) {
        XrActionCreateInfo actionGazeInfo{XR_TYPE_ACTION_CREATE_INFO};
        strcpy_s(actionGazeInfo.actionName, "user_intent");
        actionGazeInfo.actionType = XR_ACTION_TYPE_POSE_INPUT;
        strcpy_s(actionGazeInfo.localizedActionName, "User Intent");
        XRCHECK(xrCreateAction(gameplayActionSet, &actionGazeInfo, &userIntentAction));

        XrPath gazePosePath;
        XRCHECK(xrStringToPath(instance, "/user/eyes_ext/input/gaze_ext/pose", &gazePosePath));

        XrActionSuggestedBinding binding;
        binding.action = userIntentAction;
        binding.binding = gazePosePath;

        gazeBindings.push_back(binding);

        XrActionSpaceCreateInfo createActionSpaceInfo{XR_TYPE_ACTION_SPACE_CREATE_INFO};
        createActionSpaceInfo.action = userIntentAction;
        createActionSpaceInfo.poseInActionSpace.orientation.w = 1.0f;
        XRCHECK(xrCreateActionSpace(session, &createActionSpaceInfo, &gazeActionSpace));

        XrReferenceSpaceCreateInfo createReferenceSpaceInfo{XR_TYPE_REFERENCE_SPACE_CREATE_INFO};
        createReferenceSpaceInfo.referenceSpaceType = XR_REFERENCE_SPACE_TYPE_LOCAL;
        createReferenceSpaceInfo.poseInReferenceSpace.orientation.w = 1.0f;
        XRCHECK(xrCreateReferenceSpace(session, &createReferenceSpaceInfo, &localReferenceSpace));
    }

#pragma region Foveated Rendering
    XrSpace renderGazeSpace;
    XrSpace viewSpace;

    if (foveationSupported) {
        XrReferenceSpaceCreateInfo createViewSpaceInfo{XR_TYPE_REFERENCE_SPACE_CREATE_INFO};
        createViewSpaceInfo.referenceSpaceType = XR_REFERENCE_SPACE_TYPE_VIEW;
        createViewSpaceInfo.poseInReferenceSpace.orientation.w = 1.0f;
        XRCHECK(xrCreateReferenceSpace(session, &createViewSpaceInfo, &viewSpace));

        XrReferenceSpaceCreateInfo createReferenceSpaceInfo{XR_TYPE_REFERENCE_SPACE_CREATE_INFO};
        createReferenceSpaceInfo.referenceSpaceType = XR_REFERENCE_SPACE_TYPE_COMBINED_EYE_VARJO;
        createReferenceSpaceInfo.poseInReferenceSpace.orientation.w = 1.0f;
        XRCHECK(xrCreateReferenceSpace(session, &createReferenceSpaceInfo, &renderGazeSpace));
    }

#pragma endregion

    if (!gazeBindings.empty()) {
        // Create suggested bindings
        XrPath eyeGazeInteractionProfilePath;
        XRCHECK(xrStringToPath(instance, "/interaction_profiles/ext/eye_gaze_interaction", &eyeGazeInteractionProfilePath));

        XrInteractionProfileSuggestedBinding suggestedGazeBindings{XR_TYPE_INTERACTION_PROFILE_SUGGESTED_BINDING};
        suggestedGazeBindings.interactionProfile = eyeGazeInteractionProfilePath;
        suggestedGazeBindings.suggestedBindings = gazeBindings.data();
        suggestedGazeBindings.countSuggestedBindings = static_cast<uint32_t>(gazeBindings.size());
        XRCHECK(xrSuggestInteractionProfileBindings(instance, &suggestedGazeBindings));
    }

    std::vector<XrAction> booleanActions{aAction, bAction, triggerAction, trackClickAction, trackTouchAction};
    // clang-format off
    std::unordered_map<XrAction, std::string> actionNames{
        {aAction, "knuckles A button | Vive Grip button"},
        {bAction, "knuckles B button | Vive Menu button"},
        {gripAction, "Grip"},
        {triggerAction, "Trigger"},
        {triggerValueAction, "Trigger Value"},
        {thumbstickXAction, "knuckles Thumbstick X | Vive Trackpad X"},
        {thumbstickYAction, "knuckles Thumbstick Y | Vive Trackpad Y"},
        {thumbstickAction, "knuckles Thumbstick XY | Vive Trackpad XY"},
        {trackClickAction, "Vive Trackpad Click"},
        {trackTouchAction, "Vive Trackpad Touch"},
    };
    // clang-format on
    std::vector<XrAction> vectorActions{triggerValueAction, thumbstickXAction, thumbstickYAction};
    std::vector<XrAction> vector2fActions{thumbstickAction};

    XrSessionBeginInfo sessionBeginInfo{XR_TYPE_SESSION_BEGIN_INFO};
    sessionBeginInfo.primaryViewConfigurationType = viewConfigType;
    res = xrBeginSession(session, &sessionBeginInfo);
    XRCHECK(res);

    XrSessionActionSetsAttachInfo attachInfo{XR_TYPE_SESSION_ACTION_SETS_ATTACH_INFO, nullptr, 1, &gameplayActionSet};
    XRCHECK(xrAttachSessionActionSets(session, &attachInfo));

    if (environmentDepthEstimationEnabled) {
        XRCHECK(xrSetEnvironmentDepthEstimationVARJO(session, XR_TRUE));
    }

    XrTime previousDisplayTime = -1;
    if (stillScene) {
        renderer->update(1.0f);
    }

    while (!gotKey()) {
        if (!renderer->showWindow()) break;

        XrEventDataBuffer buffer{};
        while (xrPollEvent(instance, &buffer) == XR_SUCCESS) {
            if (buffer.type == XR_TYPE_EVENT_DATA_INTERACTION_PROFILE_CHANGED) {
                XrInteractionProfileState interationProfileState{XR_TYPE_INTERACTION_PROFILE_STATE};
                if (XR_SUCCEEDED(xrGetCurrentInteractionProfile(session, leftSubactionPath, &interationProfileState))) {
                    std::array<char, 256> profileName{};
                    uint32_t profileNameLength = 0;
                    xrPathToString(instance, interationProfileState.interactionProfile, static_cast<uint32_t>(std::size(profileName)), &profileNameLength,
                        profileName.data());
                    printf("Interaction profile changed: %s", profileName.data());
                }
            }
        }

        XrFrameWaitInfo frameWaitInfo{XR_TYPE_FRAME_WAIT_INFO};
        XrFrameState frameState{XR_TYPE_FRAME_STATE};
        XRCHECK(xrWaitFrame(session, &frameWaitInfo, &frameState));

        if (!isVisible(sessionState)) {
            continue;
        }

#pragma region Actions
        XrActiveActionSet activeActionSet;
        activeActionSet.actionSet = gameplayActionSet;
        activeActionSet.subactionPath = XR_NULL_PATH;
        XrActionsSyncInfo syncInfo{XR_TYPE_ACTIONS_SYNC_INFO};
        syncInfo.activeActionSets = &activeActionSet;
        syncInfo.countActiveActionSets = 1;
        XRCHECK(xrSyncActions(session, &syncInfo));

        XrActionStatePose pose{XR_TYPE_ACTION_STATE_POSE};
        XrActionStateGetInfo getInfo{XR_TYPE_ACTION_STATE_GET_INFO};
        getInfo.action = controllersAction;
        getInfo.subactionPath = leftSubactionPath;
        XRCHECK(xrGetActionStatePose(session, &getInfo, &pose));

        pose.isActive = false;
        getInfo.subactionPath = rightSubactionPath;
        XRCHECK(xrGetActionStatePose(session, &getInfo, &pose));
        if (pose.isActive) {
            XrSpaceLocation spaceRelation{XR_TYPE_SPACE_LOCATION};
            XRCHECK(xrLocateSpace(rightControllerSpace, appSpace, frameState.predictedDisplayTime, &spaceRelation));
            // auto p = spaceRelation.pose.position;
            // printf("%f, %f, %f\n", p.x, p.y, p.z);
        }

        XrActionStateFloat v{XR_TYPE_ACTION_STATE_FLOAT};
        getInfo.action = gripAction;
        getInfo.subactionPath = rightSubactionPath;
        XRCHECK(xrGetActionStateFloat(session, &getInfo, &v));
        if (v.changedSinceLastSync) {
            printf("right grip: %f\n", v.currentState);
        }

        for (auto sideAndSubactionPath : {std::make_pair("left", leftSubactionPath), std::make_pair("right", rightSubactionPath)}) {
            const auto side = sideAndSubactionPath.first;
            const auto subactionPath = sideAndSubactionPath.second;

            for (auto action : booleanActions) {
                XrActionStateBoolean booleanState{XR_TYPE_ACTION_STATE_BOOLEAN};
                getInfo.action = action;
                getInfo.subactionPath = subactionPath;
                XRCHECK(xrGetActionStateBoolean(session, &getInfo, &booleanState));
                if (booleanState.isActive && booleanState.changedSinceLastSync) {
                    printf("%s %s %s\n", side, actionNames[action].c_str(), booleanState.currentState ? "pressed" : "released");
                }
            }

            for (auto action : vectorActions) {
                XrActionStateFloat vectorState{XR_TYPE_ACTION_STATE_FLOAT};
                getInfo.action = action;
                getInfo.subactionPath = subactionPath;
                XRCHECK(xrGetActionStateFloat(session, &getInfo, &vectorState));
                if (vectorState.isActive && vectorState.changedSinceLastSync) {
                    printf("%s %s %f\n", side, actionNames[action].c_str(), vectorState.currentState);
                }
            }

            for (auto action : vector2fActions) {
                getInfo.action = action;
                getInfo.subactionPath = subactionPath;
                XrActionStateVector2f vector2fState{XR_TYPE_ACTION_STATE_VECTOR2F};
                XRCHECK(xrGetActionStateVector2f(session, &getInfo, &vector2fState));
                if (vector2fState.isActive && vector2fState.changedSinceLastSync) {
                    printf("%s %s : X:%f, Y:%f\n", side, actionNames[action].c_str(), vector2fState.currentState.x, vector2fState.currentState.y);
                }
            }
        }

        XrActionStateBoolean b{XR_TYPE_ACTION_STATE_BOOLEAN};
        getInfo.action = aAction;
        getInfo.subactionPath = rightSubactionPath;
        xrGetActionStateBoolean(session, &getInfo, &b);
        if (b.isActive && b.changedSinceLastSync) {
            printf("right a: %d\n", b.currentState);
        }

#pragma endregion

        if (gazeInteractionProperties.supportsEyeGazeInteraction) {
            XrActionStateGetInfo getGazeInfo{XR_TYPE_ACTION_STATE_GET_INFO};
            getGazeInfo.action = userIntentAction;
            XrActionStatePose gazePose{XR_TYPE_ACTION_STATE_POSE};
            XRCHECK(xrGetActionStatePose(session, &getGazeInfo, &gazePose));
            if (gazePose.isActive) {
                XrEyeGazeSampleTimeEXT eyeGazeSampleTime{XR_TYPE_EYE_GAZE_SAMPLE_TIME_EXT};
                XrSpaceLocation gazeLocation{XR_TYPE_SPACE_LOCATION, &eyeGazeSampleTime};
                xrLocateSpace(gazeActionSpace, localReferenceSpace, frameState.predictedDisplayTime, &gazeLocation);
                printf("%f, %f, %f, %f\n", gazeLocation.pose.orientation.w, gazeLocation.pose.orientation.x, gazeLocation.pose.orientation.y,
                    gazeLocation.pose.orientation.z);
            }
        }

        if (!stillScene && previousDisplayTime > 0) {
            static constexpr float NANOSECONDS_IN_SECOND = 1'000'000'000;
            float deltaTime = (frameState.predictedDisplayTime - previousDisplayTime) / NANOSECONDS_IN_SECOND;
            renderer->update(deltaTime);
        }

        previousDisplayTime = frameState.predictedDisplayTime;

        bool foveationActive = false;
        if (foveationSupported) {
            XrSpaceLocation renderGazeLocation{XR_TYPE_SPACE_LOCATION};
            XRCHECK(xrLocateSpace(renderGazeSpace, viewSpace, frameState.predictedDisplayTime, &renderGazeLocation));

            foveationActive = (renderGazeLocation.locationFlags & XR_SPACE_LOCATION_ORIENTATION_TRACKED_BIT) != 0;
        }

        XrFrameBeginInfo frameBeginInfo{XR_TYPE_FRAME_BEGIN_INFO};
        XRCHECK(xrBeginFrame(session, &frameBeginInfo));

        std::vector<XrCompositionLayerProjectionView> projectionLayerViews;
        std::vector<XrCompositionLayerDepthInfoKHR> depthLayerViews;

        XrViewState viewState{XR_TYPE_VIEW_STATE};
        uint32_t viewCapacityInput = static_cast<uint32_t>(views.size());
        uint32_t viewCountOutput;

        XrViewLocateFoveatedRenderingVARJO foveatedLocateInfo{XR_TYPE_VIEW_LOCATE_FOVEATED_RENDERING_VARJO, nullptr, foveationActive};
        XrViewLocateInfo viewLocateInfo{XR_TYPE_VIEW_LOCATE_INFO, &foveatedLocateInfo};
        viewLocateInfo.displayTime = frameState.predictedDisplayTime;
        viewLocateInfo.space = appSpace;

        XrResult res = xrLocateViews(session, &viewLocateInfo, &viewState, viewCapacityInput, &viewCountOutput, views.data());
        XRCHECK(res);
        if (XR_UNQUALIFIED_SUCCESS(res)) {
            projectionLayerViews.resize(viewCountOutput);
            depthLayerViews.resize(viewCountOutput);

            for (uint32_t i = 0; i < viewCountOutput; i++) {
                const SwapChain colorSwapchain = swapchains[i];
                XrSwapchainImageAcquireInfo acquireInfo{XR_TYPE_SWAPCHAIN_IMAGE_ACQUIRE_INFO};
                uint32_t swapchainImageIndex;
                XRCHECK(xrAcquireSwapchainImage(colorSwapchain.handle, &acquireInfo, &swapchainImageIndex));

                XrSwapchainImageWaitInfo waitInfo{XR_TYPE_SWAPCHAIN_IMAGE_WAIT_INFO};
                waitInfo.timeout = XR_INFINITE_DURATION;
                XRCHECK(xrWaitSwapchainImage(colorSwapchain.handle, &waitInfo));

                projectionLayerViews[i] = {XR_TYPE_COMPOSITION_LAYER_PROJECTION_VIEW};
                projectionLayerViews[i].pose = views[i].pose;
                projectionLayerViews[i].fov = views[i].fov;
                projectionLayerViews[i].subImage.swapchain = colorSwapchain.handle;
                projectionLayerViews[i].subImage.imageRect.offset = {0, 0};
                const uint32_t subImageWidth = foveationActive ? foveatedViews[i].recommendedImageRectWidth : configViews[i].recommendedImageRectWidth;
                const uint32_t subImageHeight = foveationActive ? foveatedViews[i].recommendedImageRectHeight : configViews[i].recommendedImageRectHeight;
                projectionLayerViews[i].subImage.imageRect.extent = {static_cast<int32_t>(subImageWidth), static_cast<int32_t>(subImageHeight)};

                SwapChain depthSwapchain{};
                uint32_t depthSwapchainImageIndex = 0;
                if (depthExtEnabled) {
                    depthSwapchain = depthSwapchains[i];
                    depthLayerViews[i] = {XR_TYPE_COMPOSITION_LAYER_DEPTH_INFO_KHR};
                    depthLayerViews[i].subImage = projectionLayerViews[i].subImage;
                    depthLayerViews[i].subImage.swapchain = depthSwapchain.handle;
                    depthLayerViews[i].minDepth = useOpengl ? -1.0f : 0.0f;
                    depthLayerViews[i].maxDepth = 1.0f;
                    depthLayerViews[i].nearZ = 0.1f;
                    depthLayerViews[i].farZ = 100.0f;

                    projectionLayerViews[i].next = &depthLayerViews[i];

                    XRCHECK(xrAcquireSwapchainImage(depthSwapchain.handle, &acquireInfo, &depthSwapchainImageIndex));
                    XRCHECK(xrWaitSwapchainImage(depthSwapchain.handle, &waitInfo));
                }

                renderer->render(colorSwapchain.handle, depthSwapchain.handle, swapchainImageIndex, depthSwapchainImageIndex, projectionLayerViews[i]);

                if (i == 3) {
                    renderer->copyForWindow(colorSwapchain.handle, swapchainImageIndex);
                }

                XrSwapchainImageReleaseInfo releaseInfo{XR_TYPE_SWAPCHAIN_IMAGE_RELEASE_INFO};
                XRCHECK(xrReleaseSwapchainImage(colorSwapchain.handle, &releaseInfo));
                if (depthExtEnabled) {
                    XRCHECK(xrReleaseSwapchainImage(depthSwapchain.handle, &releaseInfo));
                }
            }
        }

        XrCompositionLayerProjection layer{XR_TYPE_COMPOSITION_LAYER_PROJECTION};
        XrCompositionLayerDepthTestVARJO depthTest{XR_TYPE_COMPOSITION_LAYER_DEPTH_TEST_VARJO, layer.next};
        if (depthTestEnabled) {
            depthTest.depthTestRangeNearZ = depthTestRangeEnabled ? 0.1f : 0.0f;
            depthTest.depthTestRangeFarZ = depthTestRangeEnabled ? 1.75f : HUGE_VALF;

            layer.next = &depthTest;
        }
        layer.space = appSpace;
        layer.viewCount = static_cast<uint32_t>(projectionLayerViews.size());
        layer.views = projectionLayerViews.data();
        layer.layerFlags = mrEnabled ? XR_COMPOSITION_LAYER_BLEND_TEXTURE_SOURCE_ALPHA_BIT : 0;
        XrCompositionLayerBaseHeader* layers[] = {reinterpret_cast<XrCompositionLayerBaseHeader*>(&layer)};

        XrFrameEndInfo frameEndInfo{XR_TYPE_FRAME_END_INFO};
        frameEndInfo.displayTime = frameState.predictedDisplayTime;
        frameEndInfo.environmentBlendMode = environmentBlendMode;
        frameEndInfo.layerCount = 1;
        frameEndInfo.layers = layers;

        XRCHECK(xrEndFrame(session, &frameEndInfo));
    }

    XRCHECK(xrRequestExitSession(session));

    XrEventDataBuffer buffer{};
    while (xrPollEvent(instance, &buffer) == XR_SUCCESS) {
        if (buffer.type == XR_TYPE_EVENT_DATA_SESSION_STATE_CHANGED) {
            XrEventDataSessionStateChanged* sessionStateChangedEvent = reinterpret_cast<XrEventDataSessionStateChanged*>(&buffer);
            if (sessionStateChangedEvent->state == XR_SESSION_STATE_EXITING) {
                break;
            }
        }
    }

    for (size_t i = 0; i < swapchains.size(); i++) {
        xrDestroySwapchain(swapchains[i].handle);
        if (depthExtEnabled) {
            xrDestroySwapchain(depthSwapchains[i].handle);
        }
    }

    XRCHECK(xrEndSession(session));
    XRCHECK(xrDestroySession(session));
    XRCHECK(xrDestroyInstance(instance));
    return 0;
}
